#![no_std]
#![no_main]

extern crate alloc;

use blocking_network_stack::Stack;
use esp_backtrace as _;
use esp_hal::{
    clock::CpuClock,
    gpio::lp_io::{
        LowPowerInput,
        LowPowerOutput},
    load_lp_code,
    lp_core::{
        LpCore,
        LpCoreWakeupSource
    },
    main,
    rng::Rng,
    // rtc_cntl::{
    //     sleep::WakeFromLpCoreWakeupSource,
    //     Rtc
    // },
    time::{
        self
    },
    timer::timg::TimerGroup,
    uart::{
        Config,
        lp_uart::LpUart
    }
};
use esp_println::{
    print,
    println
};
use esp_wifi::{
    init,
    wifi::{
        utils::create_network_interface,
        AccessPointInfo,
        ClientConfiguration,
        Configuration,
        WifiError,
        WifiStaDevice,
    }
};
use mqtt::Mqtt;

use smoltcp::{
    iface::{
        SocketSet,
        SocketStorage
    },
    wire::{
        DhcpOption,
        Ipv4Address
    }
};

const SSID: &str = env!("SSID");
const PASSWORD: &str = env!("PASSWORD");

fn as_millis() -> u64 {
    time::now().duration_since_epoch().to_millis()
}

#[main]
fn main() -> ! {
    esp_println::logger::init_logger_from_env();
    let config = esp_hal::Config::default().with_cpu_clock(CpuClock::max());
    let peripherals = esp_hal::init(config);

    esp_alloc::heap_allocator!(72 * 1024);
    let timg0 = TimerGroup::new(peripherals.TIMG0);
    let mut rng = Rng::new(peripherals.RNG);
    let init = init(timg0.timer0, rng, peripherals.RADIO_CLK).unwrap();
    let mut wifi = peripherals.WIFI;
    let (iface, device, mut controller) =
        create_network_interface(&init, &mut wifi, WifiStaDevice).unwrap();
    controller
        .set_power_saving(esp_wifi::config::PowerSaveMode::None)
        .unwrap();
    let mut socket_set_entries: [SocketStorage; 3] = Default::default();
    let mut socket_set = SocketSet::new(&mut socket_set_entries[..]);
    let mut dhcp_socket = smoltcp::socket::dhcpv4::Socket::new();
    // we can set a hostname here (or add other DHCP options)
    dhcp_socket.set_outgoing_options(&[DhcpOption {
        kind: 12,
        data: b"esp32-c6",
    }]);
    socket_set.add(dhcp_socket);

    let stack = Stack::new(iface, device, socket_set, as_millis, rng.random());
    let client_config = Configuration::Client(ClientConfiguration {
        ssid: SSID.try_into().unwrap(),
        password: PASSWORD.try_into().unwrap(),
        ..Default::default()
    });
    let res = controller.set_configuration(&client_config);
    println!("wifi_set_configuration returned {:?}", res);

    controller.start().unwrap();
    println!("is wifi started: {:?}", controller.is_started());

    println!("Start Wifi Scan");
    let res: Result<(heapless::Vec<AccessPointInfo, 10>, usize), WifiError> = controller.scan_n();
    if let Ok((res, _count)) = res {
        for ap in res {
            println!("{:?}", ap.ssid);
        }
    }

    println!("{:?}", controller.capabilities());
    println!("wifi_connect {:?}", controller.connect());

    // wait to get connected
    println!("Wait to get connected");
    loop {
        match controller.is_connected() {
            Ok(true) => break,
            Ok(false) => {}
            Err(err) => {
                println!("{:?}", err);
                // FIXME: Error handling
                loop {}
            }
        }
    }
    println!("{:?}", controller.is_connected());

    // wait for getting an ip address
    println!("Wait to get an ip address");
    loop {
        stack.work();

        if stack.is_iface_up() {
            println!("got ip {:?}", stack.get_ip_info());
            break;
        }
    }

    let mut rx_buffer = [0u8; 1536];
    let mut tx_buffer = [0u8; 1536];
    let socket = stack.get_socket(&mut rx_buffer, &mut tx_buffer);

    let address = Ipv4Address::new(192, 168, 178, 143);
    let port: u16 = 1883;
    let mut mqtt = Mqtt::new(socket, address, port);
    match mqtt.open() {
        Ok(()) => {},
        Err(err) => panic!("unable to open mqtt connection: {:?}", err),
    }
    match mqtt.connect("stm32-c6") {
        Ok(_size) => {},
        Err(err) => panic!("unable to connect to mqtt broker: {:?}", err),
    }
    match mqtt.publish("testtopic", b"testpayload") {
        Ok(_size) => {},
        Err(err) => panic!("unable to publish to mqtt broker: {:?}", err),
    }
    mqtt.disconnect();

    let lp_pin = LowPowerOutput::new(peripherals.GPIO1);
    let uart_cfg = Config::default();
    let tx = LowPowerOutput::new(peripherals.GPIO5);
    let rx = LowPowerInput::new(peripherals.GPIO4);
    let lp_uart = LpUart::new(peripherals.LP_UART, uart_cfg, tx, rx);
    // let mut rtc = Rtc::new(peripherals.LPWR);
    // let reason = WakeFromLpCoreWakeupSource::default();
    let mut lp_core = LpCore::new(peripherals.LP_CORE);

    lp_core.stop();
    println!("lp-core -> stop");

    // load code to LP core
    let lp_core_code =
        load_lp_code!("./target/riscv32imac-unknown-none-elf/release/lp-core");

    // start LP core
    lp_core_code.run(&mut lp_core, LpCoreWakeupSource::HpCpu, lp_uart, lp_pin);
    println!("lpcore -> run");

    let mut last: u32 = 0;
    let data = (0x5000_2000) as *mut u32;
    loop {
        let read = unsafe {
            data.read_volatile()
        };
        if last != read {
            print!("Current {:x}           \u{000d}", read);
            last = read;
        }
        // FIXME...
        // println!("hp-core -> deep sleep");
        // rtc.sleep_light(&[&reason]);
    }
}

pub fn sleep_millis(delay: u32) {
    let sleep_end = as_millis() + delay as u64;
    while as_millis() < sleep_end {
        // wait
    }
}