#![no_std]

use blocking_network_stack::IoError;
use embedded_io::{
    Read,
    Write
};
use mqttrust::encoding::v4::{
    encode_slice,
    Connect,
    Packet,
    Protocol,
    Publish,
    QoS
};
use smoltcp::wire::{
    IpAddress,
    Ipv4Address
};

pub struct Mqtt <'a, 'n, D: smoltcp::phy::Device> {
    socket: blocking_network_stack::Socket<'a, 'n, D>,
    address: Ipv4Address,
    port: u16
}

impl <'n, 'a: 'n, D: smoltcp::phy::Device> Mqtt<'a, 'a, D> {
    pub fn new(
        socket: blocking_network_stack::Socket<'a, 'n, D>,
        address: Ipv4Address,
        port: u16
    ) -> Self {
        Self {
            socket,
            address,
            port
        }
    }

    pub fn open(&mut self) -> Result<(), IoError> {
        self.socket.open(IpAddress::Ipv4(self.address), self.port)
    }

    pub fn connect(&mut self, client_id: &str) -> Result<usize, IoError>{
        let connect = Packet::Connect(Connect {
            protocol: Protocol::MQTT311,
            keep_alive: 10,
            client_id,
            clean_session: true,
            last_will: None,
            username: None,
            password: None,
        });
        let mut buf = [0u8; 1024];
        let len = encode_slice(&connect, &mut buf).unwrap();
        self.socket.write(&buf[..len])
    }

    pub fn publish(&mut self, topic: &str, payload: &[u8]) -> Result<usize, IoError> {
        let packet = Packet::Publish(Publish {
            dup: false,
            qos: QoS::AtMostOnce,
            pid: None,
            retain: false,
            topic_name: topic,
            payload,
        });
        let mut buf = [0u8; 1024];
        let len = encode_slice(&packet, &mut buf).unwrap();
        let written = match self.socket.write(&buf[..len]) {
            Ok(value) => value,
            Err(err) => return core::prelude::v1::Err(err),
        };
        let _ = match self.socket.read(&mut buf) {
            Ok(value) => value,
            Err(err) => return core::prelude::v1::Err(err),
        };
        Ok(written)
    }

    pub fn disconnect(&mut self) {
        self.socket.disconnect();
    }
}
